from datetime import datetime
from dataclasses import dataclass
from math import ceil
from typing import Callable, Any
from urllib.parse import quote_plus

from sqlalchemy import func, select, Select, MappingResult
from sqlalchemy.orm.session import Session
from sqlalchemy.sql.elements import BinaryExpression


@dataclass
class ORMCheckRule:
    """规则检验"""
    errcode: Any
    condition: BinaryExpression


@dataclass
class ORMFormatRule:
    """格式化"""
    filed: str
    format: Callable


def format_datetime(dt: datetime) -> str:
    """格式化日期"""
    return dt.strftime("%Y-%m-%d %H:%M:%S")


def format_filed(data: dict, rules: list[ORMFormatRule] = []) -> dict:
    """格式化读取的字段"""
    basic_rules = [
        ORMFormatRule("create_dt", format_datetime),
        ORMFormatRule("update_dt", format_datetime),
    ]

    basic_rules.extend(rules)

    for obj in basic_rules:
        if filed := data.get(obj.filed, None):
            data[obj.filed] = obj.format(filed)

    return data


class ORM:
    @staticmethod
    def build_engine_url(dialect_dbapi: str, host: str, port: str, username: str = "", passwd: str = "", database: str = "", **kw) -> str:
        """创建sqlalchemy的engine url

        Args:
            dialect_dbapi:数据库方言和数据库API,比如'pymsql+mysql','postgresql'
            host:主机地址
            port:端口号
            username:用户名
            passwd:密码
            database:数据库名
            **kw:数据库的参数

        Return:
            str: 返回格式为:dialect+dbapi://username:passwd@host:port/database?key=value&key=value
        """
        # 拼接用户名以及密码
        username, passwd = map(quote_plus, [username, passwd])
        auth_info = f"{username}:{passwd}" if passwd else username

        # 设置数据库信息
        database = f"/{database}" if database else ""

        # 设置参数信息
        if kw:
            args = '&'.join([f"{k}={kw[k]}" for k in kw])
            param = f"?{args}"
        else:
            param = ""

        # 拼接url
        url = f"{dialect_dbapi}://{auth_info}@{host}:{port}{database}{param}"
        return url

    @staticmethod
    def mapping(db: Session, stmt) -> MappingResult:
        """ORM执行结果mapping
        """
        return db.execute(stmt).mappings()

    @staticmethod
    def all(db: Session, stmt: Select, format_rules: list[ORMFormatRule] = []) -> list[dict]:
        """获取所有数据
        """
        ds = ORM.mapping(db, stmt)
        result = [format_filed(dict(**one), format_rules) for one in ds]
        return result

    @staticmethod
    def one(db: Session, stmt: Select, format_rules: list[ORMFormatRule] = []) -> dict | None:
        """获取第一行数据
        """
        try:
            one = next(ORM.mapping(db, stmt))
            data = format_filed(dict(**one), format_rules)
        except StopIteration:
            return None

        return data

    @staticmethod
    def counts(db: Session, stmt: Select) -> int:
        """获取数据量
        """
        return db.scalar(stmt.with_only_columns(func.count("1")))

    @staticmethod
    def pagination(db: Session, stmt: Select, page_idx: int = 1, page_size: int = 10, order: list = None, format_rules: list[ORMFormatRule] = []) -> dict:
        """分页查询数据
        """
        if page_size < 1:
            page_size = 1

        match(page_idx):
            case page_idx if page_idx > 0:
                offset = (page_idx - 1) * page_size
            case _:
                offset = 0
                page_idx = 1

        total = ORM.counts(db, stmt)
        page_total = ceil(total / page_size)

        pagination = dict(page_idx=page_idx, page_size=page_size,
                          page_total=page_total, total=total)

        # 配置排序条件
        if order is not None:
            stmt = stmt.order_by(*order)

        # 配置分页条件
        stmt = stmt.offset(offset).limit(page_size)
        records = ORM.all(db, stmt, format_rules)
        data = dict(records=records, pagination=pagination)

        return data

    @staticmethod
    def check(db: Session, rules: list[ORMCheckRule], except_expression: BinaryExpression = None) -> str | int | None:
        """规则判断

        Args:
            db: 数据库会话
            rules: 检测规则
            expression: 例外表达式(比如,修改的时候不判断本条数据)

        Return:
            None:检测通过
            str | int:检测未通过,异常代码
        """
        base_stmt = select(1)

        if except_expression is not None:
            base_stmt = base_stmt.where(except_expression)

        for rule in rules:
            stmt = base_stmt.where(rule.condition)
            if ORM.counts(db, stmt) > 0:
                return rule.errcode



"""
日期:2023-12-01
内容:
- 集成钉钉和企业微信的webhook能力
参考文档:
- https://open.dingtalk.com/document/robots/custom-robot-access
- https://developer.work.weixin.qq.com/document/path/91770
"""
from base64 import b64encode
from enum import Enum
from hashlib import sha256
from hmac import new as hmac_new
from requests import session
from time import time
from urllib.parse import quote_plus


class WebHookType(int, Enum):
    DINGTALK = 0
    WECOM = 1


class WebHook:
    def __init__(self, webhook: WebHookType, url: str, sk: str = None) -> None:
        """Webhook对象

        Args:
            webhook:WebHookType,webhook的类型
            url:str,webhook的url
            sk:str,webhook的密钥,由webhook提供
        """
        self.__webhook = webhook
        self.__url = url
        self.__sk = sk

    @property
    def webhook(self):
        return self.__webhook

    @property
    def url(self):
        return self.__url

    @property
    def sk(self):
        return self.__sk

    def text(self, content: str, phone: list[str] = None, id: list[str] = None, all: bool = False) -> dict:
        """text类型,支持钉钉和企微

        Args:
            content:str,发送内容
            phone:list[str],需要@的用户手机号,比如['151****1234', '151****5678']
            id:list[str],需要@的用户ID,比如['id001', 'id002']
            all:bool,是否@ALL, 仅钉钉有效,企微请在id或phone里用'@all'来表示
        """
        result = dict(msgtype="text")
        text = dict(content=content)

        match(self.webhook):
            case WebHookType.DINGTALK:
                at = dict(atMobiles=phone, isAtAll=all, atUserIds=id)
                result["at"] = at
            case WebHookType.WECOM:
                text["mentioned_list"] = id
                text["mentioned_mobile_list"] = phone
        result["text"] = text
        return result

    def link(self, title: str, text: str, url: str, pic: str = "") -> dict:
        """link类型,仅支持钉钉

        Args:
            title:str,消息标题
            text:str,消息内容.如果太长只会展示部分
            url:str,点击消息跳转URL
            pic:str,图片URL
        """
        result = dict(msgtype="link")
        link = dict(title=title, text=text, messageUrl=url, picUrl=pic)
        result["link"] = link
        return result

    def markdown(self, text: str, title: str = "") -> dict:
        """markdown类型,支持钉钉和企微

        Args:
            text:str,markdown内容
            title:str,markdown标题,仅钉钉需要填写
        """
        result = dict(msgtype="markdown")

        match(self.webhook):
            case WebHookType.DINGTALK:
                markdown = dict(text=text, title=title)
            case WebHookType.WECOM:
                markdown = dict(content=text)

        result["markdown"] = markdown
        return result

    def send(self, data: dict) -> None:
        """发送webhook消息

        Args:
            data:dict,Webhook消息,可用类的text,markdown等方法构造
        """
        url = self.url
        sk = self.sk

        match(self.webhook):
            case WebHookType.DINGTALK:
                # 钉钉的处理
                if sk:
                    sk_enc = sk.encode()
                    timestamp = str(round(time()*1000))
                    sign_str = f"{timestamp}\n{sk}"
                    sign_str_enc = sign_str.encode()
                    hmac_code = hmac_new(sk_enc, sign_str_enc,
                                         digestmod=sha256).digest()
                    sign = quote_plus(b64encode(hmac_code))
                    url = f"{url}&timestamp={timestamp}&sign={sign}"

        # 消息发送
        with session() as s:
            result = s.post(url, json=data)

        rsp = result.json()
        if rsp["errcode"] != 0:
            raise Exception(rsp["errmsg"])


if __name__ == "__main__":
    dt = WebHook(WebHookType.DINGTALK, "url", "sk")

    data = dt.text("钉钉测试内容", phone=["151****9832"])
    data = dt.link(title="百度", text="点我可以打开baidu的地址",
                   url="https://www.baidu.com")
    data = dt.markdown("# 这是markdown的内容", title="markdown文本")
    dt.send(data)

    wb = WebHook(WebHookType.WECOM, "url")
    data = wb.text("微信测试内容", phone=["@all"])
    data = wb.markdown("# 这是markdown的内容")
    wb.send(data)



"""
整合钉钉,飞书第三方网页登录
"""
from enum import Enum
from urllib.parse import quote_plus
from requests import post, get


class AuthType(Enum):
    # 钉钉
    DINGTALK: int = 0
    # 飞书
    FEISHU: int = 1


def dingding_user_info_get(ak: str, sk: str, auth_code: str) -> dict:
    """获取钉钉的用户信息

    Args:
        ak: 钉钉应用的AppKey
        sk: 钉钉应用的SecurityKey
        auth_code: 钉钉扫码得到的auth_code

    Return:
        用户信息    
    """
    # 获取Access_token
    ACCESS_TOKEN_URL = "https://api.dingtalk.com/v1.0/oauth2/userAccessToken"
    data = dict(clientId=ak, clientSecret=sk,
                code=auth_code, grantType="authorization_code")
    with post(ACCESS_TOKEN_URL, json=data) as rsp:
        access_token = rsp.json().get("accessToken", "")

    # 获取UserInfo
    USER_INFO_URL = "https://api.dingtalk.com/v1.0/contact/users/me"
    headers = {"x-acs-dingtalk-access-token": f"{access_token}"}
    with get(USER_INFO_URL, headers=headers) as rsp:
        results = rsp.json()
    return results


def feishu_user_info_get(ak: str, sk: str, auth_code: str) -> dict:
    """获取飞书的用户信息

    Args:
        ak: 飞书应用的AppKey
        sk: 飞书应用的SecurityKey
        auth_code: 飞书扫码得到的auth_code

    Return:
        用户信息
    """
    # 1.获取app_access_token
    app_access_token_data = dict(app_id=ak, app_secret=sk)
    APP_ACCESS_TOKEN_URL = "https://open.feishu.cn/open-apis/auth/v3/app_access_token/internal"
    with post(APP_ACCESS_TOKEN_URL, json=app_access_token_data) as rsp:
        app_access_token = rsp.json().get("app_access_token", "")

    # 2.获取access_token
    ACCESS_TOKEN_URL = "https://open.feishu.cn/open-apis/authen/v1/oidc/access_token"
    access_token_headers = dict(Authorization=f"Bearer {app_access_token}")
    access_token_data = dict(
        grant_type="authorization_code", code=auth_code)
    with post(ACCESS_TOKEN_URL, headers=access_token_headers, json=access_token_data) as rsp:
        user_access_token = rsp.json()["data"]["access_token"]

    # 3.获取用户信息
    USER_INFO_URL = "https://open.feishu.cn/open-apis/authen/v1/user_info"
    user_access_token_headers = dict(
        Authorization=f"Bearer {user_access_token}")
    with get(USER_INFO_URL, headers=user_access_token_headers) as rsp:
        result = rsp.json()
    return result


class ThridAuth:
    def __init__(self, ak: str, sk: str, auth_type: AuthType) -> None:
        """三方扫码登录初始化

        Args:
            ak:str,应用的appkey
            sk:str,应用的secretkey
            auth_type:AuthType,应用类型,当前支持 钉钉:AuthType.DINGTALK,飞书:AuthType.FEISHU
        """
        self.auth_type = auth_type
        self.ak = ak
        self.sk = sk

        # 回调函数注册
        match(auth_type):
            case AuthType.DINGTALK:
                self.get_user_info_fn = dingding_user_info_get
            case AuthType.FEISHU:
                self.get_user_info_fn = feishu_user_info_get
            case _:
                raise TypeError("认证类型类型错误")

    def login_url_generate(self, redirect_url: str, state: str = None) -> str:
        """生成三方扫码登录的url地址

        Args:
            redirect_url: 扫码通过后,回调的地址
            state: 生成地址用的state

        Return:
            三方扫码登录的url地址
        """
        match(self.auth_type):
            case AuthType.DINGTALK:
                state = state or "JHU_DINGTALK"
                # 钉钉三方登录网页地址拼接
                prefix_url = "https://login.dingtalk.com/oauth2/challenge.htm"
                data = dict(
                    redirect_uri=quote_plus(redirect_url),
                    client_id=self.ak,
                    response_type="code", scope="openid", state=state, prompt="consent"
                )
            case AuthType.FEISHU:
                # 飞书三方登录网页地址拼接
                state = state or "JHU_FEISHU"
                prefix_url = "https://open.feishu.cn/open-apis/authen/v1/authorize"
                data = dict(
                    redirect_uri=quote_plus(redirect_url),
                    app_id=self.ak,
                    scope="contact:user.phone:readonly",
                    state=state
                )
            case _:
                raise TypeError("认证类型类型错误")

        query_params = '&'.join([f"{k}={data[k]}" for k in data])
        url = prefix_url+'?'+query_params
        return url

    def get_user_info(self, auth_code: str) -> dict:
        """获取三方扫码得到的用户信息

        Args:
            auth_code:str,三方登录连接返回的auth_code

        Return:
            用户的json信息
        """
        return self.get_user_info_fn(self.ak, self.sk, auth_code)


if __name__ == "__main__":
    ding_ak = "ding_ak"
    ding_sk = "ding_sk"
    d = ThridAuth(ding_ak, ding_sk, AuthType.DINGTALK)
    d.login_url_generate("http://localhost:9000/")
    # 这里的code从redirect_url返回的code中获取
    d.get_user_info("968999a21b9b38c794d230c5d80742bc")

    feishu_ak = "feishu_ak"
    feishu_sk = "feishu_sk"
    f = ThridAuth(feishu_ak, feishu_sk, AuthType.FEISHU)
    f.login_url_generate("http://localhost:9000/")
    # 这里的code从redirect_url返回的code中获取
    f.get_user_info("b8ag458bd0f54712be8c6877273ce415")



#! /usr/bin/env python3

"""
制作PIP包教程
1. 先设置一个setup文件
2. 运行setup.py sdist
3. 然后会得到一个dist文件, 这个时候就可用pip install dist/xxxxx 来安装文件了
4. 到PYPI上创建一个API Token(现在不支持user/passwd模型上传package了)
5. 在本地的创建一个$HOME/.pypirc文件
    其中:
    [pypi]
    username = __token__
    password = # 这里改成放申请下来的token
6. twine upload dist/* 可上传安装,需要输入pypi的用户名和密码
"""
from setuptools import setup, find_packages

setup(
    # 包的名称
    name="jhu",
    # 包的版本,每次更新或升级包都需要更新它
    version="1.5.6",
    # 描述
    description="jhu是一个工具包,简化或自动化的做一些任务",
    # 作者
    author="J.Hu",
    # 作者邮箱
    email="joestarhu@163.com",
    # github
    url="https://github.com/joestarhu/jhu",

    packages=find_packages(
        # 包内不需要引用的文件
        exclude=["test", "examples", "script", "tutorials"],
    ),

    classifiers=[
        "Environment :: Web Environment",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3.12"
    ],

    # 依赖的包
    install_requires=[
        # auth.py ,webhook.py
        "requests",
        # security.py
        "python-jose",
        # security.py
        "bcrypt",
        # security.py
        "pycryptodomex",
        # orm.py
        "sqlalchemy",
    ],
    zip_safe=True
)
